<?php
	// Sync protocol in a nutshell
	//
	// client sends a GET r=get, server either response with an RMC document or the 
	// empty doc if there are no documents waiting.
	//
	// Whenever the client receives a document from the server, its next request must include
	// and ack, so the client sends an GET r=ack+get&stid=<stid of doc it just received>.  
	//
	// The server will keep sending the same document until it has received an ack for that 
	// document. So the following would send the same document over and over
	//
	//	>>> GET r=get
	//	<<< <rmc stid=1">...</rmc>
	//	>>> GET r=get
	//	<<< <rmc stid=1">...</rmc>
	//	>>> GET r=get
	//	<<< <rmc stid=1">...</rmc>
	//	>>> GET r=get
	//	<<< <rmc stid=1">...</rmc>
	//
	// Only when the client acks stid=1 will it send the next document
	//
	//	>>> GET r=get
	//	<<< <rmc stid=1">...</rmc>
	//	>>> GET r=ack+get&stid=1
	//	<<< <rmc stid=2">...</rmc>
	//	>>> GET r=ack+get&stid=2
	//	<<< <rmc stid=3">...</rmc>
	//
	// If the client wants to send a document, it send a post request POST r=post with the body
	// of the request being the document being posted.  The server will respond either with
	// a document from the queue or the empty document.
	//
	// If the client has multiple documents it wants to send, it will naturally gravitate to
	// doing a ack+get loop until all documents waiting to be received are consumed then will
	// finish off all the posts requests.  So the protocol goes something like this
	//
	//	>>> POST r=post	(doc from outbox)
	//	<<< <rmc stid=10">...</rmc>
	//	>>> GET r=ack+get&stid=10
	//	<<< <rmc stid=11">...</rmc>
	//	>>> GET r=ack+get&stid=11
	//	<<< <rmc stid=12">...</rmc>
	//	>>> GET r=ack+get&stid=12
	//	<<< <rmc type="empty">
	//	>>> POST r=post	(doc from outbox)
	//	<<< <rmc type="empty">
	//	>>> POST r=post	(doc from outbox)
	//	<<< <rmc type="empty">
	//	>>> POST r=post	(doc from outbox)
	//	<<< <rmc type="empty">
	//	>>> POST r=attach (some attachment linked to one of the documents just sent)
	//	<<< <rmc type="empty">
	//
	// Note: The current protocol requires that the server alwats resonds to an r=attach
	// request with an empty document or error.
	//

	$now = time();

	// DEBUG LOG
	function DEBUG($message) {
		if (!file_exists("rmc.log")) {
			$f = fopen("rmc.log", "w");
			@chmod("rmc.log", 0666);
		} else {
			$f = fopen("rmc.log", "a");
		}
		if ($f) {
			fputs($f,$message."\n");
			fclose($f);
		} else {
			echo "<!-- $message -->";
		}
	}

	// Simulated Reliability (0-99) 0 = completely broken, 99 = 100% relaible
	$reliability = 99;

	// 0 = No simulated failures
	// 1 = 503 Service Unavailable
	// 2 = Simulated network stall before received any response
	// 3 = Simulated error saving attachment
	// 4 = Simulated protocol error (attachment response includes document, which it shouldnt)
	// 5 = Simulated network stall when authenticating
	// 6 = Simulated IP BLOCK
	// 7 = Simulated POST failure
	// 7 = Simulated registration failure (server broken)
	// 8 = XML Parse Error
	$failtype = 0;
	if ($_SERVER['REQUEST_METHOD'] != "OPTIONS") {
		if (rand(0,99) > $reliability) {
			$failtype = rand(1,9);
		}
		DEBUG("Reliability: $reliability Fail Type: $failtype");
	}

	header("Access-Control-Allow-Origin: *");
	header("Access-Control-Allow-Headers: x-requested-with, content-type");
	header("Access-Control-Expose-Headers: x-rmc-cid");

	// Empty document
	$empty = '<rmc type="empty"></rmc>';

	// Simulate service unavailable
	if ($failtype == 1) {
		DEBUG("SIMULATED: HTTP/1.0 503 Service Unavailable");
		header("HTTP/1.0 503 Service Unavailable");
		exit(0);
	}

	// Simulated IP block
	if ($failtype == 6) {
		DEBUG("BLOCKED");
		header("Content-Type: text/xml;");
		echo "<".'?xml version="1.0" encoding="ISO-8859-1"?'.">\n";
		echo '<rmc><response status="blocked" silent="true"><data>Simulated IP Block</data></response></rmc>';
		exit(0);
	}

	function openQueue($d,$mode) {
		$f = null;
		if (file_exists("queue_$d.txt")) {
			@chmod("queue_$d.txt", 0666);
			$f = fopen("queue_$d.txt",$mode);
			if (!$f) {
				DEBUG("ERROR: Can't open queue as reading for queue_$d.txt");
			}
		} else {
			DEBUG("ERROR: Queue does not exist for queue_$d.txt");
		}
		return $f;
	}

	function Q($d,$doc) {
		$f = openQueue($d,"a");
		if ($f) {
			fputs($f,$doc."\n");
			fclose($f);
		}
	}

	function removeDocFromQueue($d) {
		DEBUG("removeDocFromQueue($d)");
		$f = openQueue($d,"r");
		if ($f) {
			$fn = fopen("queue_$d.txt.new", "w");
			if ($fn) {
				$removed = false;
				while ($l = fgets($f)) {
					if (!$removed) {
						if (substr($l,0,1) != "#") { 	// ignore comment lines (consume them)
							$doc = str_replace("\n","",$l);
							$removed = true;
							// DEBUG("REMOVED: $doc");
						}
					} else {
						// DEBUG("COPY ".str_replace("\n","",$l));
						fputs($fn,$l);
					}
				}
				fclose($fn);
				fclose($f);
				if (file_exists("queue_$d.txt.new")) {
					@unlink("queue_$d.txt");
					@rename("queue_$d.txt.new", "queue_$d.txt");
					@chmod("queue_$d.txt", 0666);
				}
			}
		} 
	}

	function savePushToken($d,$token) {
		DEBUG("savePushToken($d,$token)");
		$f = fopen("token_".$d.".txt", "w");
		if ($f) {
			fputs($f,$token);
			fclose($f);
		}
	}

	function waitAck($d,$doc,$STID) {
		DEBUG("waitAck($d,$doc,$STID)");
		$s = fopen("queue_".$d."_sent.txt","w");
		if ($s) {
			fputs($s,"$doc $STID\n");
			fclose($s);
			@chmod("queue_".$d."_sent.txt", 0666);
		}
	}

	function ack($d,$STID,$error) {
		DEBUG("ack($d,$STID,$error)");
		if (file_exists("queue_".$d."_sent.txt")) {
			$doc = file_get_contents("queue_".$d."_sent.txt");
			$doc = explode(" ", str_replace("\n","",$doc));
			$stid = $doc[1];
			$doc = $doc[0];
			if ($STID == "" || $stid == $STID) {
				DEBUG("(ACK) REMOVE DOC $doc STID $stid FROM QUEUE");
				@unlink("queue_".$d."_sent.txt");
			} else {
				DEBUG("(ACK) IGNORE DUPLICATE ACK '$STID'");
			}
		} else {
			DEBUG("IGNORE DUPLICATE ACK (there is no _sent.txt file)");
		}
	}

	function getDocFromQueue($d,&$resend) {
		// if waiting for an ack for a document, and we get a request but no ack yet, resend the document
		$s = @fopen("queue_".$d."_sent.txt","r");
		if ($s) {
			$resend = true;
			$doc = explode(" ",fgets($s));
			$doc = $doc[0];
			DEBUG("RESEND: $doc");
		} else {
			// get next document from the queue
			$resend = false;
			$f = openQueue($d,"r");
			if ($f) {
				$doc = "";
				while ($l = fgets($f)) {
					if (!$doc) {
						if (substr($l,0,1) != "#") { 	// ignore comment lines (consume them)
							$doc = $l; 
							break;
						}
					}
				}
				fclose($f);
			}
			else {
				return null;		// unknown device
			}
		}
		$doc = str_replace("\n","",$doc);
		DEBUG("getDocFromQueue(): $doc");
		return $doc;
	}

	function get($d) {

		global $empty, $now;

		// read the next document from the queue
		$doc = getDocFromQueue($d,$resend);
		if (is_null($doc)) {
			DEBUG("ERROR: Unknown Device $d");
			return '<rmc type="_control" stid="dont-ack"><action type="reset" reason="Unknown Device '.$d.'"/></rmc>';
		}

		$STID = crc32($doc) ^ $now;
		$SMID = microtime(true);

		// remember document just sent
		// even if this is a resend because the stid has changed
		// but don't remove doc from queue if resend
		waitAck($d, $doc, $STID);
		if (!$resend) {
			removeDocFromQueue($d);
		}

		// if not EOF return document contents
		if ($doc != "") {
			DEBUG("READQUEUE: filename=$doc");
			if (file_exists("./docs/$doc")) {
				if (strstr($doc,".rmc") || strstr($doc,".pda")) {
					DEBUG("READFILE ./docs/$doc");
					$xml = @file_get_contents("./docs/$doc");
					if ($xml == "") {
						// file is not (yet at least) available
						DEBUG("$doc is EMPTY");
						// return error 
						return '<rmc><response status="-1"><data>'.$doc.' is empty (permissions?)</data></response></rmc>';
						// just return EOF
						// return $empty;
					}
					# strip off headers
					$h = "";
					DEBUG("PARSE HEADERS: [len=" & strlen($xml) & "] " & substr($xml,0,5));
					while (strlen($xml) && substr($xml,0,5) != "<rmc ") {
						$i = strpos($xml,"\n");
						DEBUG("NEWLINE $i");
						$h = $h . substr($xml,0,$i+1);
						DEBUG("HEADERS $h");
						$xml = substr($xml,$i+1);
						DEBUG("PARSE HEADERS: [len=" & strlen($xml) & "] " && substr($xml,0,5));
					}
					if ($xml == "") {
						return '<rmc><response status="-1"><data>cant find rmc node in docs/'.$doc.'</data></response></rmc>';
					}
					# Insert IDs
					$xml = '<rmc stid="'.$STID.'" smid="'.$SMID.'"'.substr($xml,4);
					if ($doc == "register.rmc") {
						$xml = str_replace('{DEVICE_ID}',$d,$xml);
					}
					DEBUG("stid=$STID smid=$SMID");
					# Return XML with headers
					return $h."<!-- doc = $doc -->\n".$xml;
				} else {
					// non-RMC document, send as raw data
					return "cid:$doc:".@file_get_contents("./docs/$doc");
				}
			} 
			DEBUG("ERROR: Can't open requested document (invalid queue entry)");
			return '<rmc><response status="-1"><data>cant open docs/'.$doc.'</data></response></rmc>';
		}

		// blank line = simulate EOF
		DEBUG("READQUEUE: empty");
		ack($d,"","");				 // need to ack the EOF ourselves, because it won't be acked by the client
		return $empty;
	}

	function get_register_rmc($d) {
		global $now;
		$SMID = microtime(true);
		$STID = crc32("register.rmc") ^ $now;
		$XML = str_replace('{DEVICE_ID}',$d,@file_get_contents("./docs/register.rmc"));
		if ($XML == "") {
			$XML = $empty;
		} else {
			$XML = '<rmc stid="'.$STID.'" smid="'.$SMID.'"'.substr($XML,4);
			$XML = "<!-- doc = register.rmc -->\n".$XML;
			copy("queue.empty", "queue_$d.txt");
			waitAck($d,"register.rmc",$STID);
		}
		return $XML;
	}

	date_default_timezone_set('Europe/London');

	if ($_SERVER['REQUEST_METHOD'] != "OPTIONS") {
		$r = $_GET["r"];

		// handle upload from test harness (method of adding documents into the queue)
		if ($r == "upload") {
			// handle document upload
			if ($_FILES["file"]["error"] > 0) {
				echo "Error: " . $_FILES["file"]["error"] . "<br />";
				exit(0);
			}

			$d = $_POST["id"];
			if (!$d) {
				echo "Error: missinge device id<br/>";
				exit(0);
			}

			echo "Upload: " . $_FILES["file"]["name"] . "<br />";
			echo "Type: " . $_FILES["file"]["type"] . "<br />";
			echo "Size: " . ($_FILES["file"]["size"] / 1024) . " Kb<br />";
			echo "Stored in: " . $_FILES["file"]["tmp_name"] . "<br/>";

			$FN=$_FILES["file"]["name"];
			$T=$_FILES["file"]["tmp_name"];

			// move the file into the queue
			if (move_uploaded_file ($_FILES["file"]["tmp_name"], "docs/".$FN)) {
				echo "Moved to: docs/".$FN."<br/>";
				$f = fopen("queue_$d.txt","a");
				if ($f) {
					fputs($f,$FN."\n");
					fclose($f);
					?>
					<p id="status">Document added to device queue for device <?php echo $d; ?>.
					<br/>Redirecting to upload form in <span id="time">5</span>...</p>
					<script>var cd = 5; window.setInterval(function() { time.innerHTML = --cd; if (cd == 0) history.back(); },1000)</script>
					<?php
				}
			}
			exit(0);
		}

		DEBUG("\n=========================================================================== ".date("Y/m/d H:i:s"));
		$d = @$_GET["deviceID"];
		DEBUG("PROCESS REQUEST [deviceID=$d]: r=$r");

		// simulate a network stall getting a response from the server
		if ($failtype == 2) {
			DEBUG("********** SIMULATED NETWORK OUTAGE (SLEEPING for 320 seconds) **********");
			sleep(320);					
			exit(0);
		}

		if ($d == "") {
			$d = @$_GET["testID"];
			if ($d == "") {
				$d = $now; $d = substr($d,2);
				DEBUG("\nNEW DEVICE REGISTRATION (id=$d)");
			} else {
				DEBUG("\nNEW TEST DEVICE REGISTRATION (id=$d)");
			}
			$XML = get_register_rmc($d);
		} else {
			switch($r) {
			case "":
				DEBUG("ERROR: r param is missing");
				$XML = '<rmc><response status="-1"><data>r param missing</data></response></rmc>';
				break;
			case "ack":
				DEBUG("ERROR: ack sent on it's own");
				$XML = '<rmc><response status="-1"><data>ack sent on its own</data></response></rmc>';
				break;
			case "attach":
				$outboxID = $_GET['ctid'];
				$docID = $_GET['docID'];
				$prefix = $_GET['prefix'];
				$attID = $_GET['attID'];
				$type = $_GET['type'];
				$name = $_GET['name'];
				$body = @file_get_contents('php://input');
				$filename = "IN/".str_replace("/","-",$outboxID."-att-".$docID."-".$attID."-".$name);
				switch($type) {
				case "image/jpeg":
				case "image/jpg":
				case "image/png":
					break;
				default:
					$filename = $filename.".bin";
					break;
				}
				DEBUG("ATTACHMENT: outboxID=".$outboxID." [docID=".$docID." prefix=".$prefix." attID=".$attID."]");
				DEBUG("POST: length=".strlen($body)." (binary data) data:$type;base64,".substr(base64_encode($body),0,50)." ..."); 
				$wrote = @file_put_contents($filename, $body, 0);
				DEBUG("PUT_CONTENTS $wrote bytes to $filename");
				// Saving the attachment is assumed to have worked here in the reference server,
				// if it didnt, then an error document would be sent, if it does work, either an
				// empty document can be sent, or we could process the queue, though the likelyhood
				// of there being a document waiting (having just finished a sync) is slim.
				if ($failtype == 3) {		// simulate error saving attachment
					DEBUG("SIMULATED attachment failure");
					$XML = '<rmc><response status="-1"><data>(simulated) attachment could not be saved</data></response></rmc>';
				} else {
					// http://bugzilla.internal.redskyit.com/show_bug.cgi?id=56894
					// Protocol says, always respond with empty document.  Client is not currently coded to properly
					// deal with receiving a document at this point.  See Bug 56894 for specifics.
					if ($failtype == 4) {
						DEBUG("SIMULATED protocol error");
						$XML = get($d);					// every once in a while, simulate protocol error
					} else {
						$XML = $empty;
					}
				}
				break;
			case "post":
				if ($failtype == 7) {
					DEBUG("POST: Simulated post error");
					$XML = '<rmc><response status="-1"><data>(simulated) Failed to execute: blah blah</data></response></rmc>';
				} else {
					$body = @file_get_contents('php://input');
					DEBUG("POST: length=".strlen($body));
					$xml = simplexml_load_string($body);
					if ($xml) {
						$type = $xml->attributes()->type;
						DEBUG("doc type: ".$type."\n");
						$ctid = $xml->ctid;
						$id = $xml->id;
						$filename = "IN/".str_replace("/","-","$ctid-$id-$type").".xml";
						$wrote = @file_put_contents($filename, $body, 0);
						DEBUG("PUT_CONTENTS $wrote bytes to $filename");
						switch($type) {
						case "register":
							$p = strpos($body,"id=\"operative\"");
							$who = substr($body,$p+14);
							$p = strpos($who,">");
							$who = substr($who,$p+1);
							$p = strpos($who,"<");
							$who = strtoupper(substr($who,0,$p));
							DEBUG("WHO: $who");
							// very crude company and operative validation
							if ($who == "ABBE01") {
								copy("queue.newdev", "queue_$d.txt");
								if ($failtype == 8) {
									DEBUG("Simulate broken registration");
									unlink("queue_$d.txt");
								}
								$XML = get($d);
							} else if ($who == "TEST") {
								// regression tests
								copy("queue.newdev-$d", "queue_$d.txt");
								$XML = get($d);
							} else if (file_exists("queue.newdev-$who")) {
								copy("queue.newdev-$who", "queue_$d.txt");
								$XML = get($d);
							} else {
								// registration failed (invalid company or operator)
								// send registration back
								$XML = get_register_rmc($d);
								Q($d,"regerror.rmc");
							}
							break;
						case "_notify":
							switch($xml->reason) {
							case "push-token":
								$token = $xml->payload->token;
								DEBUG("push-token $token");
								savePushToken($d, $token);
								break;
							case "push-ack":
								$ackid = $xml->payload->id;
								DEBUG("push-ack $ackid");
								break;
							}
							$XML = get($d);
							break;
						default:
							// assumed to have worked
							$XML = get($d);
						}
					} else {
						DEBUG($body);
						// Invalid document
						$XML = '<rmc><response status="-1"><data>XML posted to the server could not be parsed</data></response></rmc>';
					}
				}
				break;
			case "ack+get":
				// remove document from the queue
				DEBUG("ACK deviceId=$d STID=".$_GET["stid"]." ERROR=".$_GET["error"]);
				ack($d,$_GET["stid"],$_GET["error"]);
				// then fall through and do a get
			case "get":
				$XML = get($d);
				break;
			case "auth":
				DEBUG("AUTHENTICATE:-");
				// simulate a network stall getting a response from the server
				if ($failtype == 5) {
					DEBUG("********** SIMULATED NETWORK OUTAGE **********");
					sleep(31);					
					exit(0);
				}
				$body = @file_get_contents('php://input');
				DEBUG($body);
				if (strpos($body,' password="demo"') === false || strpos($body, ' user="demo"') === false) {
					DEBUG("AUTHENTICATION FAILED");
					$XML = '<rmc type="_auth" status="4" reason="User Not Found"/>';
				} else {
					DEBUG("AUTHENTICATION SUCCESSFUL");
					$XML = '<rmc type="_auth" status="0"/>';
				}
				break;
			default:
				DEBUG("ERROR: r param invalid");
				$XML = '<rmc><response status="-1"><data>r param invalid '.$r.'</data></response></rmc>';
				break;
			}
		}

		// check for raw document
		if (substr($XML,0,4) == "cid:") {
			$XML = substr($XML,4);
			$doc = substr($XML,0,strpos($XML,":"));
			$XML = substr($XML,strlen($doc)+1);

			// $doc = ID
			// $XML = raw data

			if (strstr($doc,".png")) {
				$ct = "image/png";
				$type = "image";
			} else if (strstr($doc,".jpg") || strstr($doc,".jpeg")) {
				$ct = "image/jpeg";
				$type = "image";
			} else if (strstr($doc,".widget") || strstr($doc,".widget.html")) {
				$ct = "application/x-rmc-widget;";
				$type = "widget";
			} else {
				$ct = "application/x-rmc-resource;";
				$type = "resource";
			} 

			DEBUG("Content-Type: $ct");

			$STID = crc32($doc) ^ $now;

			$cid = preg_replace("/.*\//","",$doc);

			if (rand(0,99) >= 80) {
				// send as base64 content encoded stream with HTTP headers
				DEBUG("SEND: [len=".strlen($XML)."] RESOURCE CONTENT cid=$cid stid=$STID");
				header("X-RMC-CID: $cid");
				header("X-RMC-STID: $STID");
				header("Content-Type: $ct;");
				header("Content-Transfer-Encoding: base64");
				echo base64_encode($XML);
			} else {
				// send as base64 encoded content RMC document
				DEBUG("SEND: [len=".strlen($XML)."] RESOURCE DOCUMENT cid=$cid stid=$STID");
				header("Content-Type: text/xml;charset=iso-8859-2");
				echo "<".'?xml version="1.0" encoding="iso-8859-2"?'.">\n"
						.'<rmc type="'.$type.'" mimetype="'.$ct.'" cid="'.$cid.'" stid="'.$STID.'" encoding="base64">'
						.base64_encode($XML)
						.'</rmc>';
			}
		} else {
			if (substr($XML,0,5) != "<"."?xml") {
				$XML = "<".'?xml version="1.0" encoding="ISO-8859-1"?'.">\n".$XML;
			}
			DEBUG("SEND: [len=".strlen($XML)."] $XML");
			header("Content-Type: text/xml;charset=UTF-8");
			if (strpos(strtolower($XML),' encoding="utf-8"') > -1) {
				header("Content-Encoding: UTF-8");
			} else {
				header("Content-Encoding: ISO-8859-1");
			}
			if ($failtype == 1) echo 'simulated parse error';
			echo $XML;
		}
	}
?>
